from __future__ import annotations
from dataclasses import dataclass
from ...services.locale.time_pattern import TimePattern
from .enums.price_type import PriceType

@dataclass
class TieredPrice:
    lower_peak_limit: float
    value: float

    def __post_init__(self):
        assert self.value >= 0
        assert self.lower_peak_limit >= 0


@dataclass
class SeasonedPrice:
    validity: TimePattern
    value: float

    def __post_init__(self):
        assert self.value >= 0


@dataclass
class LocalePrice:
    price_type: PriceType
    value: float | list[TieredPrice] | list[SeasonedPrice]
    currency:str

    def __post_init__(self):
        match self.price_type:
            case PriceType.Static:
                assert isinstance(self.value, (float,int))
            case PriceType.Seasoned:
                assert type(self.value) == list
                assert isinstance(self.value[0], SeasonedPrice)
            case PriceType.Tiered:
                assert type(self.value) == list
                assert isinstance(self.value[0], TieredPrice)
        assert len(self.currency) > 0

    @property
    def price(self) -> float:
        match self.price_type:
            case PriceType.Static:
                return self.value
            case PriceType.Seasoned:
                return self._get_price_seasoned()
            case PriceType.Tiered:
                return self._get_price_tiered()

    def is_equal(self, other_currency: str) -> bool:
        """Use this method if necessary to test against the el-price currency"""
        return self.currency.lower == other_currency.lower

    def _get_price_seasoned(self) -> float:
        t: SeasonedPrice
        for t in self.value:
            if t.validity.free_charge():
                return t.value        

    def _get_price_tiered(self):
        # t: TieredPrice
        # for t in self.value:
        pass
    


from enum import Enum

class CautionHourType(Enum):
    SUAVE = "Suave"
    INTERMEDIATE = "Intermediate"
    AGGRESSIVE = "Aggressive"

    @staticmethod
    def get_num_value(type_state):
        values = {
            CautionHourType.SUAVE.value: 0.75,
            CautionHourType.INTERMEDIATE.value: 0.5,
            CautionHourType.AGGRESSIVE.value: 0.4
            }

        if isinstance(type_state, str):
            tt = type_state
        elif isinstance(type_state, CautionHourType):
            tt = type_state.value

        return values[tt]



from dataclasses import dataclass, field
from typing import List, Dict
from .hourobject import HourObject
import logging

_LOGGER = logging.getLogger(__name__)


@dataclass(frozen=False)
class HoursModel:
    non_hours: List[int] = field(default_factory=lambda : [])
    caution_hours: List[int] = field(default_factory=lambda : [])
    dynamic_caution_hours: Dict[int, float] = field(default_factory=lambda : {})
    hours_today: HourObject = field(default_factory=lambda : HourObject([],[],dict()))
    hours_tomorrow: HourObject = field(default_factory=lambda : HourObject([],[],dict()))
    offset_dict: Dict[Dict[str,float], Dict[str, float]] = field(default_factory=lambda: {})

    def update_non_hours(
        self, 
        hour:int
        ) -> None:
        ret = []
        ret.extend(h for h in self.hours_today.nh if h >= hour)
        ret.extend(h for h in self.hours_tomorrow.nh if h < hour)
        self.non_hours = ret
    
    def update_caution_hours(
        self, 
        hour:int
        ) -> None:
        ret = []
        ret.extend(h for h in self.hours_today.ch if h >= hour)
        ret.extend(h for h in self.hours_tomorrow.ch if h < hour)
        self.caution_hours = ret

    def update_dynanmic_caution_hours(
        self, 
        hour:int
        ) -> None:
        ret = {}
        ret.update({k: v for k, v in self.hours_today.dyn_ch.items() if k >= hour and k not in self.hours_today.nh})
        ret.update({k: v for k, v in self.hours_tomorrow.dyn_ch.items() if k < hour and k not in self.hours_tomorrow.nh})
        self.dynamic_caution_hours = ret

    def update_offset_dict(self) -> None:
        ret = {}
        ret['today'] = self.hours_today.offset_dict
        ret['tomorrow'] = self.hours_tomorrow.offset_dict
        self.offset_dict = ret

@dataclass(frozen=False)
class HourSelectionOptions:
    cautionhour_type: float = 0
    top_price: float = 0
    min_price: float = 0
    absolute_top_price: float = field(init=False)

    def __post_init__(self):
        self.absolute_top_price = self.set_absolute_top_price(self.top_price, self.min_price)

    def set_absolute_top_price(self, top, min) -> float:
        if not self.validate_top_min_prices(top, min):
            top = 0
            HourSelectionOptions.min_price = 0
            _LOGGER.warning(f"Setting top-price and min-price to zero because of min-price being larger than top-price. Please fix in options. top:{top} min:{min}")
        if top is None:
            return float("inf")
        if top <= 0:
            return float("inf")
        return float(top)

    def validate_top_min_prices(self, top, min) -> bool:
        if any(
            [top == 0, min == 0]
        ):  
            return True
        return top > min
        

@dataclass(frozen=False)
class HourSelectionModel:
    prices_today: List[float] = field(default_factory=lambda : [])
    prices_tomorrow: List[float] = field(default_factory=lambda : [])
    adjusted_average: float = None
    current_peak: float = 0.0
    hours: HoursModel = HoursModel()
    options: HourSelectionOptions = HourSelectionOptions

    def __post_init__(self):
        self.validate()

    def validate(self):
        assert 0 < self.options.cautionhour_type <= 1
        assert isinstance(self.prices_today, list)
        assert isinstance(self.prices_tomorrow, list)

        if isinstance(self.adjusted_average, (int, float)):
            assert self.adjusted_average >= 0
        else:
            assert self.adjusted_average is None
        


from dataclasses import dataclass

@dataclass(frozen=False)
class ChargerOptions:
    powerswitch_controls_charging: bool = False #todo: rename this member to something else. It currently conflicts with the servicecall-option named the same (but with different outcome)
    charger_is_outlet: bool = False
    ampmeter_is_attribute: bool = False
    powermeter_factor: int = 1
    authentication_required: bool = False


from ....models.locale.queryservice_model import QueryServiceModel, Group, DatePartModel
from ....models.locale.enums.datepart_datetype import DatePartDateType
from ....models.locale.enums.datepart_modeltype import DatePartModelType
from .const import (
    QUERYTYPE_AVERAGEOFTHREEHOURS_MON_FRI_07_19,
    QUERYTYPE_HIGHLOAD,
    QUERYTYPE_MAX_NOV_MAR_MON_FRI_06_22,
    QUERYTYPE_SOLLENTUNA,
    QUERYTYPE_BASICMAX_MON_FRI_07_17_DEC_MAR_ELSE_REGULAR
    )
from .queryservice import Dividents

QUERYSETS = {
    QUERYTYPE_SOLLENTUNA: QueryServiceModel(
    [Group(divident=Dividents.AND, dateparts=[
        DatePartModel(type=DatePartModelType.GreaterOrEqual, dttype=DatePartDateType.Hour, values=[7]),
        DatePartModel(type=DatePartModelType.LessOrEqual, dttype=DatePartDateType.Hour, values=[18]),
        DatePartModel(type=DatePartModelType.LessOrEqual, dttype=DatePartDateType.Weekday, values=[4])
        ]
        )
    ]
    ),
    QUERYTYPE_MAX_NOV_MAR_MON_FRI_06_22: QueryServiceModel(
    [Group(divident=Dividents.AND, dateparts=[
            DatePartModel(type=DatePartModelType.GreaterOrEqual, dttype=DatePartDateType.Hour, values=[6]),
            DatePartModel(type=DatePartModelType.LessOrEqual, dttype=DatePartDateType.Hour, values=[22]),
            DatePartModel(type=DatePartModelType.LessOrEqual, dttype=DatePartDateType.Weekday, values=[4]),
            DatePartModel(type=DatePartModelType.In, dttype=DatePartDateType.Month, values=[11, 12, 1, 2, 3])
        ]
        )
    ]
    ),
    QUERYTYPE_AVERAGEOFTHREEHOURS_MON_FRI_07_19: QueryServiceModel(
        [Group(divident=Dividents.AND, dateparts=[
            DatePartModel(type=DatePartModelType.In, dttype=DatePartDateType.Hour, values=[7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18]),
            DatePartModel(type=DatePartModelType.LessOrEqual, dttype=DatePartDateType.Weekday, values=[4])
        ]
        )
    ]
    ),
    QUERYTYPE_BASICMAX_MON_FRI_07_17_DEC_MAR_ELSE_REGULAR: QueryServiceModel(
        [Group(divident=Dividents.AND, dateparts=[
            DatePartModel(type=DatePartModelType.In, dttype=DatePartDateType.Hour, values=[7, 8, 9, 10, 11, 12, 13, 14, 15, 16]),
            DatePartModel(type=DatePartModelType.LessOrEqual, dttype=DatePartDateType.Weekday, values=[4]),
            DatePartModel(type=DatePartModelType.In, dttype=DatePartDateType.Month, values=[12, 1, 2, 3])
        ]
        ),
        Group(divident=Dividents.OR, dateparts=[
            DatePartModel(type=DatePartModelType.In, dttype=DatePartDateType.Month, values=[4, 5, 6, 7, 8, 9, 10, 11])
            ])
    ]
    ),
    QUERYTYPE_HIGHLOAD: QueryServiceModel(
        [Group(divident=Dividents.AND, dateparts=[
            DatePartModel(type=DatePartModelType.In, dttype=DatePartDateType.Hour, values=[8, 9, 10, 11, 12, 13, 14, 15, 16,17,18]),
            DatePartModel(type=DatePartModelType.LessOrEqual, dttype=DatePartDateType.Weekday, values=[4])
        ]
        )
    ]
    )
}


from .price_aware_hours import PriceAwareHours
from .regular_hours import RegularHours

class HourselectionFactory:
    @staticmethod
    def create(hub):
        if hub.options.price.price_aware is False:
            return RegularHours(hub)
        return PriceAwareHours(hub)


import pytest

from ..services.chargertype.chargertype_base import ChargerBase
from ..services.chargertype.servicecalls import ServiceCalls, ServiceCallsDTO, ServiceCallsOptions
from ..models.chargertype.calltype import CallType
from ..services.chargertype.const import (
    PARAMS, DOMAIN
)
from ..models.chargertype.calltype_enum import CallTypes

MOCKDOMAIN = "peaqchargerbox"

TESTON = CallType("on")
TESTOFF = CallType("off")
TESTPAUSE = CallType("pause")
TESTRESUME = CallType("resume")
TESTUPDATECURRENT = CallType("set_max_limit", {"testparam1": "param1", "testparam2": "param2"})

def test_get_pause_and_resume_from_blank():
    model = ServiceCallsDTO(TESTON, TESTOFF)
    opt = ServiceCallsOptions(allowupdatecurrent=False, update_current_on_termination=True)
    s = ServiceCalls(MOCKDOMAIN, model, opt)
    pause = s.get_call(CallTypes.Pause)
    resume = s.get_call(CallTypes.Resume)

    assert pause["domain"] == resume["domain"] == MOCKDOMAIN
    assert pause["params"] == TESTOFF.params
    assert pause[CallTypes.Pause] == TESTOFF.call
    assert resume["params"] == TESTON.params
    assert resume[CallTypes.Resume] == TESTON.call

def test_call_updatecurrent_invalid():
    model = ServiceCallsDTO(TESTON, TESTOFF, TESTUPDATECURRENT)
    opt = ServiceCallsOptions(allowupdatecurrent=False, update_current_on_termination=True)
    s = ServiceCalls(MOCKDOMAIN, model, opt)
    with pytest.raises(AttributeError):
              s.get_call(CallTypes.UpdateCurrent)
